@@ -22,6 +22,10 @@ THE SOFTWARE.
 
 /* Backend-independent encoding code */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <string.h>
 #include <strings.h>
 #include <stdlib.h>
@@ -33,6 +37,7 @@ THE SOFTWARE.
 
 #include <X11/fonts/fontenc.h>
 #include "fontencI.h"
+#include "reallocarray.h"
 
 /* Functions local to this file */
 
@@ -274,14 +279,14 @@ iso8859_7_to_unicode(unsigned isocode, void *client_data)
         isocode == 0xB7 || isocode == 0xBB || isocode == 0xBD)
         return isocode;
     else if (isocode == 0xA1)
-        return 0x02BD;
+        return 0x2018;
     else if (isocode == 0xA2)
-        return 0x02BC;
+        return 0x2019;
     else if (isocode == 0xAF)
         return 0x2015;
     else if (isocode == 0xD2)   /* unassigned */
         return 0;
-    else if (isocode >= 0xB4)
+    else if (isocode >= 0xB4 && isocode <= 0xFE)
         return isocode - 0xA0 + 0x0370;
     else
         return 0;
@@ -808,7 +813,7 @@ FontEncLoad(const char *encoding_name, const char *filename)
                 for (alias = encoding->aliases; *alias; alias++)
                     numaliases++;
             }
-            new_aliases = malloc((numaliases + 2) * sizeof(char *));
+            new_aliases = Xmallocarray(numaliases + 2, sizeof(char *));
             if (new_aliases == NULL) {
                 free(new_name);
                 return NULL;
